@@ -1,9 +1,11 @@
 from .app import IdfApp
 from .dut import IdfDut
 from .serial import IdfSerial
+from .unity_tester import CaseTester
 
 __all__ = [
     'IdfApp',
     'IdfSerial',
     'IdfDut',
+    'CaseTester',
 ]
